@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
 # License: GPLv3 Copyright: 2010, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -11,23 +10,25 @@ from contextlib import contextmanager
 from html5_parser import parse
 from lxml import html
 from qt.core import (
-    QAction, QApplication, QBrush, QByteArray, QCheckBox, QColor, QColorDialog,
-    QDialog, QDialogButtonBox, QFont, QFontInfo, QFontMetrics, QFormLayout,
-    QHBoxLayout, QIcon, QKeySequence, QLabel, QLineEdit, QMenu, QPalette,
-    QPlainTextEdit, QPushButton, QSize, QSyntaxHighlighter, Qt, QTabWidget,
-    QTextBlockFormat, QTextCharFormat, QTextCursor, QTextEdit, QTextFormat,
-    QTextListFormat, QToolBar, QToolButton, QUrl, QVBoxLayout, QWidget, pyqtSignal,
-    pyqtSlot
+    QAction, QApplication, QBrush, QByteArray, QCheckBox, QColor, QColorDialog, QDialog,
+    QDialogButtonBox, QFont, QFontInfo, QFontMetrics, QFormLayout, QHBoxLayout, QIcon,
+    QKeySequence, QLabel, QLineEdit, QMenu, QPalette, QPlainTextEdit, QPushButton,
+    QSize, QSyntaxHighlighter, Qt, QTabWidget, QTextBlockFormat, QTextCharFormat,
+    QTextCursor, QTextEdit, QTextFormat, QTextListFormat, QTimer, QToolButton, QUrl,
+    QVBoxLayout, QWidget, pyqtSignal, pyqtSlot,
 )
 
 from calibre import xml_replace_entities
+from calibre.db.constants import DATA_DIR_NAME
 from calibre.ebooks.chardet import xml_to_unicode
 from calibre.gui2 import (
-    NO_URL_FORMATTING, choose_files, error_dialog, gprefs, is_dark_theme
+    NO_URL_FORMATTING, choose_dir, choose_files, error_dialog, gprefs, is_dark_theme,
 )
-from calibre.gui2.book_details import css
+from calibre.gui2.book_details import resolved_css
+from calibre.gui2.flow_toolbar import create_flow_toolbar
 from calibre.gui2.widgets import LineEditECM
 from calibre.gui2.widgets2 import to_plain_text
+from calibre.startup import connect_lambda
 from calibre.utils.cleantext import clean_xml_chars
 from calibre.utils.config import tweaks
 from calibre.utils.imghdr import what
@@ -56,10 +57,12 @@ def lift_styles(tag, style_map):
     has_text = bool(tag.text)
     child_styles = []
     for child in tag.iterchildren('*'):
-        if child.tail:
+        if child.tail and child.tail.strip():
             has_text = True
         style = style_map[child]
         child_styles.append(style)
+        if not child.text and len(child) == 0:
+            continue
         if common_props is None:
             common_props = style.copy()
         else:
@@ -76,7 +79,7 @@ def lift_styles(tag, style_map):
         if lifted_props:
             for style in child_styles:
                 for k in lifted_props:
-                    del style[k]
+                    style.pop(k, None)
 
 
 def filter_qt_styles(style):
@@ -104,13 +107,13 @@ def remove_heading_font_styles(tag, style):
     expected_size = (None, 'xx-large', 'x-large', 'large', None, 'small', 'x-small')[lvl]
     if style.get('font-size', 1) == expected_size:
         del style['font-size']
-    if style.get('font-weight') == '600':
+    if style.get('font-weight') in ('0', '600', '700', 'bold'):
         del style['font-weight']
 
 
 def use_implicit_styling_for_span(span, style):
     is_italic = style.get('font-style') == 'italic'
-    is_bold = style.get('font-weight') == '600'
+    is_bold = style.get('font-weight') in ('600', '700', 'bold')
     if is_italic and not is_bold:
         del style['font-style']
         span.tag = 'em'
@@ -140,7 +143,7 @@ def use_implicit_styling_for_a(a, style_map):
 def merge_contiguous_links(root):
     all_hrefs = set(root.xpath('//a/@href'))
     for href in all_hrefs:
-        tags = root.xpath('//a[@href="{}"]'.format(href))
+        tags = root.xpath(f'//a[@href="{href}"]')
         processed = set()
 
         def insert_tag(parent, child):
@@ -184,6 +187,7 @@ def cleanup_qt_markup(root):
     style_map = defaultdict(dict)
     for tag in root.xpath('//*[@style]'):
         style_map[tag] = parse_style(tag.get('style'))
+    convert_anchors_to_ids(root)
     block_tags = root.xpath('//body/*')
     for tag in block_tags:
         lift_styles(tag, style_map)
@@ -207,19 +211,58 @@ def cleanup_qt_markup(root):
                     remove_zero_indents(ts)
     for style in itervalues(style_map):
         filter_qt_styles(style)
+        fw = style.get('font-weight')
+        if fw in ('600', '700'):
+            style['font-weight'] = 'bold'
     for tag, style in iteritems(style_map):
         if style:
-            tag.set('style', '; '.join('{}: {}'.format(k, v) for k, v in iteritems(style)))
+            tag.set('style', '; '.join(f'{k}: {v}' for k, v in iteritems(style)))
         else:
             tag.attrib.pop('style', None)
     for span in root.xpath('//span[not(@style)]'):
         lift(span)
 
     merge_contiguous_links(root)
-    convert_anchors_to_ids(root)
 # }}}
 
 
+def fix_html(original_html, original_txt, remove_comments=True):
+    raw = original_html
+    raw = xml_to_unicode(raw, strip_encoding_pats=True, resolve_entities=True)[0]
+    if remove_comments:
+        comments_pat = re.compile(r'<!--.*?-->', re.DOTALL)
+        raw = comments_pat.sub('', raw)
+    if not original_txt and '<img' not in raw.lower():
+        return ''
+
+    try:
+        root = parse(raw, maybe_xhtml=False, sanitize_names=True)
+    except Exception:
+        root = parse(clean_xml_chars(raw), maybe_xhtml=False, sanitize_names=True)
+    if root.xpath('//meta[@name="calibre-dont-sanitize"]'):
+        # Bypass cleanup if special meta tag exists
+        return original_html
+
+    try:
+        cleanup_qt_markup(root)
+    except Exception:
+        import traceback
+        traceback.print_exc()
+    elems = []
+    for body in root.xpath('//body'):
+        if body.text:
+            elems.append(body.text)
+        elems += [html.tostring(x, encoding='unicode') for x in body if
+            x.tag not in ('script', 'style')]
+
+    if len(elems) > 1:
+        ans = '<div>%s</div>'%(''.join(elems))
+    else:
+        ans = ''.join(elems)
+        if not ans.startswith('<'):
+            ans = '<p>%s</p>'%ans
+    return xml_replace_entities(ans)
+
 class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
     data_changed = pyqtSignal()
@@ -245,27 +288,28 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
     def __init__(self, parent=None):
         QTextEdit.__init__(self, parent)
         self.setTabChangesFocus(True)
-        self.document().setDefaultStyleSheet(css() + '\n\nli { margin-top: 0.5ex; margin-bottom: 0.5ex; }')
+        self.document().setDefaultStyleSheet(resolved_css() + '\n\nli { margin-top: 0.5ex; margin-bottom: 0.5ex; }')
         font = self.font()
         f = QFontInfo(font)
         delta = tweaks['change_book_details_font_size_by'] + 1
         if delta:
-            font.setPixelSize(f.pixelSize() + delta)
+            font.setPixelSize(int(f.pixelSize() + delta))
             self.setFont(font)
         f = QFontMetrics(self.font())
         self.em_size = f.horizontalAdvance('m')
         self.base_url = None
         self._parent = weakref.ref(parent)
-        self.comments_pat = re.compile(r'<!--.*?-->', re.DOTALL)
+        self.shortcut_map = {}
 
         def r(name, icon, text, checkable=False, shortcut=None):
-            ac = QAction(QIcon(I(icon + '.png')), text, self)
+            ac = QAction(QIcon.ic(icon + '.png'), text, self)
             ac.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
             if checkable:
                 ac.setCheckable(checkable)
             setattr(self, 'action_'+name, ac)
             ac.triggered.connect(getattr(self, 'do_' + name))
             if shortcut is not None:
+                self.shortcut_map[shortcut] = ac
                 sc = shortcut if isinstance(shortcut, QKeySequence) else QKeySequence(shortcut)
                 ac.setShortcut(sc)
                 ac.setToolTip(text + f' [{sc.toString(QKeySequence.SequenceFormat.NativeText)}]')
@@ -289,7 +333,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         r('remove_format', 'edit-clear', _('Remove formatting'))
         r('copy', 'edit-copy', _('Copy'), shortcut=QKeySequence.StandardKey.Copy)
         r('paste', 'edit-paste', _('Paste'), shortcut=QKeySequence.StandardKey.Paste)
-        r('paste_and_match_style', 'edit-paste', _('Paste and match style'))
+        r('paste_and_match_style', 'edit-paste', _('Paste and match style'), shortcut=QKeySequence('ctrl+shift+v', QKeySequence.SequenceFormat.PortableText))
         r('cut', 'edit-cut', _('Cut'), shortcut=QKeySequence.StandardKey.Cut)
         r('indent', 'format-indent-more', _('Increase indentation'))
         r('outdent', 'format-indent-less', _('Decrease indentation'))
@@ -302,7 +346,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         r('insert_hr', 'format-text-hr', _('Insert separator'),)
         r('clear', 'trash', _('Clear'))
 
-        self.action_block_style = QAction(QIcon(I('format-text-heading.png')),
+        self.action_block_style = QAction(QIcon.ic('format-text-heading.png'),
                 _('Style text block'), self)
         self.action_block_style.setToolTip(
                 _('Style the selected text block'))
@@ -372,7 +416,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             if bf.leftMargin() == bf.rightMargin() and bf.leftMargin() > 0:
                 name = 'blockquote'
         else:
-            name = 'h{}'.format(lvl)
+            name = f'h{lvl}'
         for ac in self.block_style_actions:
             ac.setChecked(ac.block_name == name)
 
@@ -488,6 +532,13 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             c.setCharFormat(QTextCharFormat())
         self.update_cursor_position_actions()
 
+    def keyPressEvent(self, ev):
+        for sc, ac in self.shortcut_map.items():
+            if isinstance(sc, QKeySequence.StandardKey) and ev.matches(sc):
+                ac.trigger()
+                return
+        return super().keyPressEvent(ev)
+
     def do_copy(self):
         self.copy()
         self.focus_self()
@@ -535,7 +586,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             cf = QTextCharFormat()
             bcf = c.blockCharFormat()
             lvl = self.level_for_block_type(name)
-            wt = QFont.Weight.Bold if lvl else None
+            wt = QFont.Weight.Bold if lvl else QFont.Weight.Normal
             adjust = (0, 3, 2, 1, 0, -1, -1)[lvl]
             pos = None
             if not c.hasSelection():
@@ -560,12 +611,10 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             bf.setLeftMargin(hmargin), bf.setRightMargin(hmargin)
             bf.setTopMargin(tmargin), bf.setBottomMargin(bmargin)
             bf.setHeadingLevel(lvl)
-            if adjust:
-                bcf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
-                cf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
-            if wt:
-                bcf.setProperty(QTextFormat.Property.FontWeight, wt)
-                cf.setProperty(QTextFormat.Property.FontWeight, wt)
+            bcf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
+            cf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
+            bcf.setFontWeight(wt)
+            cf.setFontWeight(wt)
             c.setBlockCharFormat(bcf)
             c.mergeCharFormat(cf)
             c.mergeBlockFormat(bf)
@@ -602,7 +651,10 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             return
         url = self.parse_link(link)
         if url.isValid():
-            url = str(url.toString(NO_URL_FORMATTING))
+            if url.isLocalFile() and not os.path.isabs(url.toLocalFile()):
+                url = url.toLocalFile()
+            else:
+                url = url.toString(NO_URL_FORMATTING)
             self.focus_self()
             with self.editing_cursor() as c:
                 if is_image:
@@ -628,7 +680,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
         else:
             error_dialog(self, _('Invalid URL'),
-                         _('The url %r is invalid') % link, show=True)
+                         _('The URL %r is invalid') % link, show=True)
 
     def ask_link(self):
 
@@ -654,24 +706,46 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         d.treat_as_image = QCheckBox(d)
         d.setMinimumWidth(600)
         d.bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok|QDialogButtonBox.StandardButton.Cancel)
-        d.br = b = QPushButton(_('&Browse'))
-        b.setIcon(QIcon(I('document_open.png')))
+        d.br = b = QPushButton(_('&File'))
+        base = os.path.dirname(self.base_url.toLocalFile()) if self.base_url else os.getcwd()
+        data_path = os.path.join(base, DATA_DIR_NAME)
+        if self.base_url:
+            os.makedirs(data_path, exist_ok=True)
 
-        def cf():
+        def cf(data_dir=False):
             filetypes = []
             if d.treat_as_image.isChecked():
                 filetypes = [(_('Images'), 'png jpeg jpg gif'.split())]
-            files = choose_files(d, 'select link file', _('Choose file'), filetypes, select_only_single_file=True)
+            if data_dir:
+                files = choose_files(
+                    d, 'select link file', _('Choose file'), filetypes, select_only_single_file=True, no_save_dir=True,
+                    default_dir=data_path)
+            else:
+                files = choose_files(d, 'select link file', _('Choose file'), filetypes, select_only_single_file=True)
             if files:
                 path = files[0]
                 d.url.setText(path)
                 if path and os.path.exists(path):
-                    with lopen(path, 'rb') as f:
+                    with open(path, 'rb') as f:
                         q = what(f)
-                    is_image = q in {'jpeg', 'png', 'gif'}
+                    is_image = q in {'jpeg', 'png', 'gif', 'webp'}
                     d.treat_as_image.setChecked(is_image)
+                if data_dir:
+                    path = os.path.relpath(path, base)
+                    d.url.setText(path)
+        b.clicked.connect(lambda: cf())
+        d.brdf = b = QPushButton(_('&Data file'))
+        b.clicked.connect(lambda: cf(True))
+        b.setToolTip(_('A relative link to a data file associated with this book'))
+        if not os.path.exists(data_path):
+            b.setVisible(False)
+        d.brd = b = QPushButton(_('F&older'))
+        def cd():
+            path = choose_dir(d, 'select link folder', _('Choose folder'))
+            if path:
+                d.url.setText(path)
+        b.clicked.connect(cd)
 
-        b.clicked.connect(cf)
         d.la = la = QLabel(_(
             'Enter a URL. If you check the "Treat the URL as an image" box '
             'then the URL will be added as an image reference instead of as '
@@ -685,13 +759,15 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         l.addRow(_('Enter &URL:'), d.url)
         l.addRow(_('Treat the URL as an &image'), d.treat_as_image)
         l.addRow(_('Enter &name (optional):'), d.name)
-        l.addRow(_('Choose a file on your computer:'), d.br)
+        h = QHBoxLayout()
+        h.addWidget(d.br), h.addWidget(d.brdf), h.addWidget(d.brd)
+        l.addRow(_('Choose a file on your computer:'), h)
         l.addRow(d.bb)
         d.bb.accepted.connect(d.accept)
         d.bb.rejected.connect(d.reject)
         d.resize(d.sizeHint())
         link, name, is_image = None, None, False
-        if d.exec_() == QDialog.DialogCode.Accepted:
+        if d.exec() == QDialog.DialogCode.Accepted:
             link, name = str(d.url.text()).strip(), str(d.name.text()).strip()
             is_image = d.treat_as_image.isChecked()
         return link, name, is_image
@@ -705,6 +781,14 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             url = QUrl(link, QUrl.ParsingMode.TolerantMode)
             if url.isValid():
                 return url
+        else:
+            if self.base_url:
+                base = os.path.dirname(self.base_url.toLocalFile())
+            else:
+                base = os.getcwd()
+            candidate = os.path.join(base, link)
+            if os.path.exists(candidate):
+                return QUrl.fromLocalFile(link)
         if os.path.exists(link):
             return QUrl.fromLocalFile(link)
 
@@ -724,40 +808,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
     @property
     def html(self):
-        raw = original_html = self.toHtml()
-        check = self.toPlainText().strip()
-        raw = xml_to_unicode(raw, strip_encoding_pats=True, resolve_entities=True)[0]
-        raw = self.comments_pat.sub('', raw)
-        if not check and '<img' not in raw.lower():
-            return ''
-
-        try:
-            root = parse(raw, maybe_xhtml=False, sanitize_names=True)
-        except Exception:
-            root = parse(clean_xml_chars(raw), maybe_xhtml=False, sanitize_names=True)
-        if root.xpath('//meta[@name="calibre-dont-sanitize"]'):
-            # Bypass cleanup if special meta tag exists
-            return original_html
-
-        try:
-            cleanup_qt_markup(root)
-        except Exception:
-            import traceback
-            traceback.print_exc()
-        elems = []
-        for body in root.xpath('//body'):
-            if body.text:
-                elems.append(body.text)
-            elems += [html.tostring(x, encoding='unicode') for x in body if
-                x.tag not in ('script', 'style')]
-
-        if len(elems) > 1:
-            ans = '<div>%s</div>'%(''.join(elems))
-        else:
-            ans = ''.join(elems)
-            if not ans.startswith('<'):
-                ans = '<p>%s</p>'%ans
-        return xml_replace_entities(ans)
+        return fix_html(self.toHtml(), self.toPlainText().strip())
 
     @html.setter
     def html(self, val):
@@ -774,7 +825,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             if qurl.isLocalFile():
                 path = qurl.toLocalFile()
                 try:
-                    with lopen(path, 'rb') as f:
+                    with open(path, 'rb') as f:
                         data = f.read()
                 except OSError:
                     if path.rpartition('.')[-1].lower() in {'jpg', 'jpeg', 'gif', 'png', 'bmp', 'webp'}:
@@ -799,20 +850,41 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
     def text(self):
         return self.textCursor().selectedText()
+    selectedText = text
 
     def setText(self, text):
         with self.editing_cursor() as c:
             c.insertText(text)
+    insert = setText
+
+    def hasSelectedText(self):
+        c = self.textCursor()
+        return c.hasSelection()
+
+    def createMimeDataFromSelection(self):
+        ans = super().createMimeDataFromSelection()
+        html, txt = ans.html(), ans.text()
+        html = fix_html(html, txt, remove_comments=False)
+        # Qt has a bug where copying from the start of a paragraph does not
+        # include the paragraph definition in the fragment. Try to fix that
+        # by moving the StartFragment comment to before the paragraph when
+        # the selection starts at the start of a block
+        c = self.textCursor()
+        c2 = QTextCursor(c)
+        c2.setPosition(c.selectionStart())
+        if c2.atBlockStart():
+            html = re.sub(r'(<p.*?>)(<!--StartFragment-->)', r'\2\1', html, count=1)
+        ans.setHtml(html)
+        return ans
 
     def contextMenuEvent(self, ev):
-        menu = self.createStandardContextMenu()
-        for action in menu.actions():
-            parts = action.text().split('\t')
-            if len(parts) == 2 and QKeySequence(QKeySequence.StandardKey.Paste).toString(QKeySequence.SequenceFormat.NativeText) in parts[-1]:
-                menu.insertAction(action, self.action_paste_and_match_style)
-                break
-        else:
-            menu.addAction(self.action_paste_and_match_style)
+        menu = QMenu(self)
+        for ac in 'undo redo -- cut copy paste paste_and_match_style -- select_all'.split():
+            if ac == '--':
+                menu.addSeparator()
+            else:
+                ac = getattr(self, 'action_' + ac)
+                menu.addAction(ac)
         st = self.text()
         m = QMenu(_('Fonts'))
         m.addAction(self.action_bold), m.addAction(self.action_italic), m.addAction(self.action_underline)
@@ -832,7 +904,7 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         am.addAction(self.action_background)
         am.addAction(self.action_color)
         menu.addAction(_('Smarten punctuation'), parent.smarten_punctuation)
-        menu.exec_(ev.globalPos())
+        menu.exec(ev.globalPos())
 
 # }}}
 
@@ -1069,12 +1141,7 @@ class Editor(QWidget):  # {{{
     def __init__(self, parent=None, one_line_toolbar=False, toolbar_prefs_name=None):
         QWidget.__init__(self, parent)
         self.toolbar_prefs_name = toolbar_prefs_name or self.toolbar_prefs_name
-        self.toolbar1 = QToolBar(self)
-        self.toolbar2 = QToolBar(self)
-        self.toolbar3 = QToolBar(self)
-        for i in range(1, 4):
-            t = getattr(self, 'toolbar%d'%i)
-            t.setIconSize(QSize(18, 18))
+        self.toolbar = create_flow_toolbar(self, restrict_to_single_line=one_line_toolbar, icon_size=18)
         self.editor = EditorWidget(self)
         self.editor.data_changed.connect(self.data_changed)
         self.set_base_url = self.editor.set_base_url
@@ -1083,6 +1150,7 @@ class Editor(QWidget):  # {{{
         self.tabs.setTabPosition(QTabWidget.TabPosition.South)
         self.wyswyg = QWidget(self.tabs)
         self.code_edit = QPlainTextEdit(self.tabs)
+        self.code_edit.setTabChangesFocus(True)
         self.source_dirty = False
         self.wyswyg_dirty = True
 
@@ -1090,14 +1158,8 @@ class Editor(QWidget):  # {{{
         self.wyswyg.layout = l = QVBoxLayout(self.wyswyg)
         self.setLayout(self._layout)
         l.setContentsMargins(0, 0, 0, 0)
-        if one_line_toolbar:
-            tb = QHBoxLayout()
-            l.addLayout(tb)
-        else:
-            tb = l
-        tb.addWidget(self.toolbar1)
-        tb.addWidget(self.toolbar2)
-        tb.addWidget(self.toolbar3)
+
+        l.addWidget(self.toolbar)
         l.addWidget(self.editor)
         self._layout.addWidget(self.tabs)
         self.tabs.addTab(self.wyswyg, _('&Normal view'))
@@ -1110,53 +1172,47 @@ class Editor(QWidget):  # {{{
             if hidden:
                 self.hide_toolbars()
 
-        # toolbar1 {{{
-        self.toolbar1.addAction(self.editor.action_undo)
-        self.toolbar1.addAction(self.editor.action_redo)
-        self.toolbar1.addAction(self.editor.action_select_all)
-        self.toolbar1.addAction(self.editor.action_remove_format)
-        self.toolbar1.addAction(self.editor.action_clear)
-        self.toolbar1.addSeparator()
+        self.toolbar.add_action(self.editor.action_undo)
+        self.toolbar.add_action(self.editor.action_redo)
+        self.toolbar.add_action(self.editor.action_select_all)
+        self.toolbar.add_action(self.editor.action_remove_format)
+        self.toolbar.add_action(self.editor.action_clear)
+        self.toolbar.add_separator()
 
         for x in ('copy', 'cut', 'paste'):
             ac = getattr(self.editor, 'action_'+x)
-            self.toolbar1.addAction(ac)
+            self.toolbar.add_action(ac)
 
-        self.toolbar1.addSeparator()
-        self.toolbar1.addAction(self.editor.action_background)
-        # }}}
+        self.toolbar.add_separator()
+        self.toolbar.add_action(self.editor.action_background)
+        self.toolbar.add_action(self.editor.action_color)
+        self.toolbar.add_separator()
 
-        # toolbar2 {{{
         for x in ('', 'un'):
             ac = getattr(self.editor, 'action_%sordered_list'%x)
-            self.toolbar2.addAction(ac)
-        self.toolbar2.addSeparator()
+            self.toolbar.add_action(ac)
+        self.toolbar.add_separator()
         for x in ('superscript', 'subscript', 'indent', 'outdent'):
-            self.toolbar2.addAction(getattr(self.editor, 'action_' + x))
+            self.toolbar.add_action(getattr(self.editor, 'action_' + x))
             if x in ('subscript', 'outdent'):
-                self.toolbar2.addSeparator()
+                self.toolbar.add_separator()
 
-        self.toolbar2.addAction(self.editor.action_block_style)
-        w = self.toolbar2.widgetForAction(self.editor.action_block_style)
-        if hasattr(w, 'setPopupMode'):
-            w.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
-        self.toolbar2.addAction(self.editor.action_insert_link)
-        self.toolbar2.addAction(self.editor.action_insert_hr)
-        # }}}
+        self.toolbar.add_action(self.editor.action_block_style, popup_mode=QToolButton.ToolButtonPopupMode.InstantPopup)
+        self.toolbar.add_action(self.editor.action_insert_link)
+        self.toolbar.add_action(self.editor.action_insert_hr)
+        self.toolbar.add_separator()
 
-        # toolbar3 {{{
         for x in ('bold', 'italic', 'underline', 'strikethrough'):
             ac = getattr(self.editor, 'action_'+x)
-            self.toolbar3.addAction(ac)
+            self.toolbar.add_action(ac)
             self.addAction(ac)
-        self.toolbar3.addSeparator()
+        self.toolbar.add_separator()
 
         for x in ('left', 'center', 'right', 'justified'):
             ac = getattr(self.editor, 'action_align_'+x)
-            self.toolbar3.addAction(ac)
-        self.toolbar3.addSeparator()
-        self.toolbar3.addAction(self.editor.action_color)
-        # }}}
+            self.toolbar.add_action(ac)
+        self.toolbar.add_separator()
+        QTimer.singleShot(0, self.toolbar.updateGeometry)
 
         self.code_edit.textChanged.connect(self.code_dirtied)
         self.editor.data_changed.connect(self.wyswyg_dirtied)
@@ -1200,14 +1256,11 @@ class Editor(QWidget):  # {{{
         self.source_dirty = True
 
     def hide_toolbars(self):
-        self.toolbar1.setVisible(False)
-        self.toolbar2.setVisible(False)
-        self.toolbar3.setVisible(False)
+        self.toolbar.setVisible(False)
 
     def show_toolbars(self):
-        self.toolbar1.setVisible(True)
-        self.toolbar2.setVisible(True)
-        self.toolbar3.setVisible(True)
+        self.toolbar.setVisible(True)
+        QTimer.singleShot(0, self.toolbar.updateGeometry)
 
     def toggle_toolbars(self):
         visible = self.toolbars_visible
@@ -1217,7 +1270,7 @@ class Editor(QWidget):  # {{{
 
     @property
     def toolbars_visible(self):
-        return self.toolbar1.isVisible() or self.toolbar2.isVisible() or self.toolbar3.isVisible()
+        return self.toolbar.isVisible()
 
     @toolbars_visible.setter
     def toolbars_visible(self, val):
@@ -1242,12 +1295,14 @@ class Editor(QWidget):  # {{{
 if __name__ == '__main__':
     from calibre.gui2 import Application
     app = Application([])
-    w = Editor()
+    w = Editor(one_line_toolbar=False)
+    w.set_base_url(QUrl.fromLocalFile(os.getcwd()))
     w.resize(800, 600)
+    w.setWindowFlag(Qt.WindowType.Dialog)
     w.show()
     w.html = '''<h1>Test Heading</h1><blockquote>Test blockquote</blockquote><p><span style="background-color: rgb(0, 255, 255); ">He hadn't
     set <u>out</u> to have an <em>affair</em>, <span style="font-style:italic; background-color:red">
     much</span> less a <s>long-term</s>, <b>devoted</b> one.</span><p>hello'''
-    w.html = '<div><p id="moo">Testing <em>a</em> link.</p><p>\xa0</p><p>ss</p></div>'
-    app.exec_()
+    w.html = '<div><p id="moo" align="justify">Testing <em>a</em> link.</p><p align="justify">\xa0</p><p align="justify">ss</p></div>'
+    app.exec()
     # print w.html
