@@ -11,11 +11,11 @@ from functools import partial
 
 from qt.core import (
     Qt, QDialog, QAbstractItemView, QTableWidgetItem, QIcon, QListWidgetItem,
-    QCoreApplication, QEvent, QObject, QApplication, pyqtSignal, QByteArray, QMenu,
+    QCoreApplication, QEvent, QObject, QApplication, pyqtSignal, QMenu,
     QShortcut, QTimer, QStyle)
 
 from calibre.customize.ui import find_plugin
-from calibre.gui2 import gprefs
+from calibre.gui2 import gprefs, error_dialog
 from calibre.gui2.dialogs.quickview_ui import Ui_Quickview
 from calibre.utils.date import timestampfromdt
 from calibre.utils.icu import sort_key
@@ -154,11 +154,10 @@ class Quickview(QDialog, Ui_Quickview):
 
     def __init__(self, gui, row, toggle_shortcut):
         self.is_pane = gprefs.get('quickview_is_pane', False)
-
         if not self.is_pane:
-            QDialog.__init__(self, gui, flags=Qt.WindowType.Widget)
+            QDialog.__init__(self, None, flags=Qt.WindowType.Window)
         else:
-            QDialog.__init__(self, gui)
+            QDialog.__init__(self, None, flags=Qt.WindowType.Dialog)
         Ui_Quickview.__init__(self)
         self.setupUi(self)
         self.isClosed = False
@@ -175,9 +174,7 @@ class Quickview(QDialog, Ui_Quickview):
             self.books_table_column_widths = \
                         gprefs.get('quickview_dialog_books_table_widths', None)
             if not self.is_pane:
-                geom = gprefs.get('quickview_dialog_geometry', None)
-                if geom:
-                    QApplication.instance().safe_restore_geometry(self, QByteArray(geom))
+                self.restore_geometry(gprefs, 'quickview_dialog_geometry')
         except:
             pass
 
@@ -289,6 +286,7 @@ class Quickview(QDialog, Ui_Quickview):
 
         self.view_icon = QIcon.ic('view.png')
         self.view_plugin = self.gui.iactions['View']
+        self.show_details_plugin = self.gui.iactions['Show Book Details']
         self.edit_metadata_icon = QIcon.ic('edit_input.png')
         self.quickview_icon = QIcon.ic('quickview.png')
         self.select_book_icon = QIcon.ic('library.png')
@@ -344,6 +342,8 @@ class Quickview(QDialog, Ui_Quickview):
         a = m.addAction(self.select_book_icon, _('Select this book in the library'),
                                 partial(self.select_book, book_id))
         a.setEnabled(book_displayed)
+        m.addAction(_('Open a locked Book details window for this book'),
+                    partial(self.show_book_details, book_id))
         m.addAction(self.search_icon, _('Find item in the library'),
                         partial(self.do_search, follow_library_view=False))
         a = m.addAction(self.edit_metadata_icon, _('Edit metadata'),
@@ -351,8 +351,10 @@ class Quickview(QDialog, Ui_Quickview):
         a.setEnabled(book_displayed)
         a = m.addAction(self.quickview_icon, _('Quickview this cell'),
                         partial(self.quickview_item, row, column))
-        a.setEnabled(self.is_category(self.column_order[column]) and
-                     book_displayed and not self.lock_qv.isChecked())
+        key = self.column_order[column]
+        a.setEnabled(self.is_category(key) and book_displayed and
+                     key in self.view.visible_columns and
+                     not self.lock_qv.isChecked())
         m.addSeparator()
         m.addAction(self.view_icon, _('Open book in the E-book viewer'),
                         partial(self.view_plugin._view_calibre_books, [book_id]))
@@ -520,9 +522,11 @@ class Quickview(QDialog, Ui_Quickview):
             self.indicate_no_items()
 
     def is_category(self, key):
-        return key is not None and (self.fm[key]['is_category'] or
+        return key is not None and (
+                     self.fm[key]['table'] is not None and
+                     (self.fm[key]['is_category'] or
                                     (self.fm[key]['datatype'] == 'composite' and
-                                     self.fm[key]['display'].get('make_category', False)))
+                                     self.fm[key]['display'].get('make_category', False))))
 
     def _refresh(self, book_id, key):
         '''
@@ -544,10 +548,7 @@ class Quickview(QDialog, Ui_Quickview):
         self.books_table.setRowCount(0)
 
         mi = self.db.new_api.get_proxy_metadata(book_id)
-        vals = mi.get(key, None)
-        if self.fm[key]['datatype'] == 'composite' and self.fm[key]['is_multiple']:
-            sep = self.fm[key]['is_multiple'].get('cache_to_list', ',')
-            vals = [v.strip() for v in vals.split(sep) if v.strip()]
+        vals = self.db.new_api.split_if_is_multiple_composite(key, mi.get(key, None))
         try:
             # Check if we are in the GridView and there are no values for the
             # selected column. In this case switch the column to 'authors'
@@ -724,7 +725,6 @@ class Quickview(QDialog, Ui_Quickview):
             self.select_book_and_qv(row, self.key_to_table_widget_column(self.current_key))
 
     def book_not_in_view_error(self):
-        from calibre.gui2 import error_dialog
         error_dialog(self, _('Quickview: Book not in library view'),
                      _('The book you selected is not currently displayed in '
                        'the library view, perhaps because of a search or a '
@@ -751,6 +751,7 @@ class Quickview(QDialog, Ui_Quickview):
             else:
                 self.quickview_item(row, self.key_to_table_widget_column(self.current_key))
         except:
+            traceback.print_exc()
             self.book_not_in_view_error()
 
     def edit_metadata(self, book_id, follow_library_view=True):
@@ -763,6 +764,12 @@ class Quickview(QDialog, Ui_Quickview):
         finally:
             self.follow_library_view = True
 
+    def show_book_details(self, book_id):
+        try:
+            self.show_details_plugin.show_book_info(book_id=book_id, locked=True)
+        finally:
+            pass
+
     def select_book(self, book_id):
         '''
         Select a book in the library view without changing the QV lists
@@ -790,6 +797,13 @@ class Quickview(QDialog, Ui_Quickview):
         if QApplication.keyboardModifiers() in (Qt.KeyboardModifier.ControlModifier, Qt.KeyboardModifier.ShiftModifier):
             self.edit_metadata(book_id)
         else:
+            if key not in self.view.visible_columns:
+                error_dialog(self, _("Quickview: Column cannot be selected"),
+                     _("The column you double-clicked, '{}', is not shown in the "
+                       "library view. The book/column cannot be selected by Quickview.").format(key),
+                     show=True,
+                     show_copy_button=False)
+                return
             self.view.select_cell(self.db.data.id_to_index(book_id),
                                   self.view.column_map.index(key))
 
@@ -824,7 +838,7 @@ class Quickview(QDialog, Ui_Quickview):
             self.books_table_column_widths.append(self.books_table.columnWidth(c))
         gprefs['quickview_dialog_books_table_widths'] = self.books_table_column_widths
         if not self.is_pane:
-            gprefs['quickview_dialog_geometry'] = bytearray(self.saveGeometry())
+            self.save_geometry(gprefs, 'quickview_dialog_geometry')
 
     def _close(self):
         self.save_state()
