@@ -4,21 +4,24 @@
 
 import weakref
 from qt.core import (
-    QApplication, QBrush, QByteArray, QCalendarWidget, QCheckBox, QColor,
-    QColorDialog, QComboBox, QDate, QDateTime, QDateTimeEdit, QDialog,
-    QDialogButtonBox, QFont, QFontInfo, QFontMetrics, QFrame, QIcon, QKeySequence,
-    QLabel, QLayout, QMenu, QMimeData, QPainter, QPalette, QPixmap, QPoint,
-    QPushButton, QRect, QScrollArea, QSize, QSizePolicy, QStyle, QStyledItemDelegate,
-    QStyleOptionToolButton, QStylePainter, Qt, QTabWidget, QTextBrowser, QTextCursor,
-    QToolButton, QUndoCommand, QUndoStack, QUrl, QWidget, pyqtSignal
+    QApplication, QBrush, QByteArray, QCalendarWidget, QCheckBox, QColor, QColorDialog,
+    QComboBox, QDate, QDateTime, QDateTimeEdit, QDialog, QDialogButtonBox, QFont,
+    QFontInfo, QFontMetrics, QFrame, QIcon, QKeySequence, QLabel, QLayout, QMenu,
+    QMimeData, QPainter, QPalette, QPixmap, QPoint, QPushButton, QRect, QScrollArea,
+    QSize, QSizePolicy, QStyle, QStyledItemDelegate, QStyleOptionToolButton,
+    QStylePainter, Qt, QTabWidget, QTextBrowser, QTextCursor, QTimer, QToolButton,
+    QUndoCommand, QUndoStack, QUrl, QWidget, pyqtSignal,
 )
 
+from calibre import prepare_string_for_xml
+from calibre.constants import builtin_colors_dark, builtin_colors_light
 from calibre.ebooks.metadata import rating_to_stars
 from calibre.gui2 import UNDEFINED_QDATETIME, gprefs, rating_font
 from calibre.gui2.complete2 import EditWithComplete, LineEdit
 from calibre.gui2.widgets import history
 from calibre.utils.config_base import tweaks
 from calibre.utils.date import UNDEFINED_DATE
+from calibre.utils.localization import _
 from polyglot.functools import lru_cache
 
 
@@ -221,23 +224,20 @@ class Dialog(QDialog):
 
         self.setup_ui()
 
-        self.resize(self.sizeHint())
-        geom = self.prefs_for_persistence.get(name + '-geometry', None)
-        if geom is not None:
-            QApplication.instance().safe_restore_geometry(self, geom)
+        self.restore_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
-            state = self.prefs_for_persistence.get(name + '-splitter-state', None)
+            state = self.prefs_for_persistence.get(self.name + '-splitter-state', None)
             if state is not None:
                 self.splitter.restoreState(state)
 
     def accept(self):
-        self.prefs_for_persistence.set(self.name + '-geometry', bytearray(self.saveGeometry()))
+        self.save_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
             self.prefs_for_persistence.set(self.name + '-splitter-state', bytearray(self.splitter.saveState()))
         QDialog.accept(self)
 
     def reject(self):
-        self.prefs_for_persistence.set(self.name + '-geometry', bytearray(self.saveGeometry()))
+        self.save_geometry(self.prefs_for_persistence, self.name + '-geometry')
         if hasattr(self, 'splitter'):
             self.prefs_for_persistence.set(self.name + '-splitter-state', bytearray(self.splitter.saveState()))
         QDialog.reject(self)
@@ -547,7 +547,7 @@ class HTMLDisplay(QTextBrowser):
 
     def setDefaultStyleSheet(self, css=''):
         self.external_css = css
-        self.document().setDefaultStyleSheet(self.default_css + self.external_css)
+        self.document().setDefaultStyleSheet(self.default_css + self.process_external_css(self.external_css))
 
     def palette_changed(self):
         app = QApplication.instance()
@@ -557,9 +557,12 @@ class HTMLDisplay(QTextBrowser):
             self.default_css = 'a { color: %s }\n\n' % col.name(QColor.NameFormat.HexRgb)
         else:
             self.default_css = ''
-        self.document().setDefaultStyleSheet(self.default_css + self.external_css)
+        self.document().setDefaultStyleSheet(self.default_css + self.process_external_css(self.external_css))
         self.setHtml(self.last_set_html)
 
+    def process_external_css(self, css):
+        return css
+
     def on_anchor_clicked(self, qurl):
         if not qurl.scheme() and qurl.hasFragment() and qurl.toString().startswith('#'):
             frag = qurl.fragment(QUrl.ComponentFormattingOption.FullyDecoded)
@@ -572,7 +575,7 @@ class HTMLDisplay(QTextBrowser):
         if qurl.isLocalFile():
             path = qurl.toLocalFile()
             try:
-                with lopen(path, 'rb') as f:
+                with open(path, 'rb') as f:
                     data = f.read()
             except OSError:
                 if path.rpartition('.')[-1].lower() in {'jpg', 'jpeg', 'gif', 'png', 'bmp', 'webp'}:
@@ -726,6 +729,63 @@ class DateTimeEdit(QDateTimeEdit):
             return QDateTimeEdit.keyPressEvent(self, ev)
 
 
+class MessagePopup(QLabel):
+
+    undo_requested = pyqtSignal(object)
+    OFFSET_FROM_TOP = 25
+
+    def __init__(self, parent):
+        QLabel.__init__(self, parent)
+        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
+        self.undo_data = None
+        if QApplication.instance().is_dark_theme:
+            c = builtin_colors_dark['green']
+        else:
+            c = builtin_colors_light['green']
+        self.color = self.palette().color(QPalette.ColorRole.WindowText).name()
+        bg = QColor(c).getRgb()
+        self.setStyleSheet(f'''QLabel {{
+            background-color: rgba({bg[0]}, {bg[1]}, {bg[2]}, 0.85);
+            border-radius: 4px;
+            color: {self.color};
+            padding: 0.5em;
+        }}'''
+        )
+        self.linkActivated.connect(self.link_activated)
+        self.close_timer = t = QTimer()
+        t.setSingleShot(True)
+        t.timeout.connect(self.hide)
+        self.setMouseTracking(True)
+        self.hide()
+
+    def mouseMoveEvent(self, ev):
+        self.close_timer.start()
+        return super().mouseMoveEvent(ev)
+
+    def link_activated(self, link):
+        self.hide()
+        if link.startswith('undo://'):
+            self.undo_requested.emit(self.undo_data)
+
+    def __call__(self, text='Testing message popup', show_undo=True, timeout=5000, has_markup=False):
+        text = '<p>' + (text if has_markup else prepare_string_for_xml(text))
+        if show_undo:
+            self.undo_data = show_undo
+            text += '\xa0\xa0<a style="text-decoration: none" href="undo://me.com">{}</a>'.format(_('Undo'))
+        text += f'\xa0\xa0<a style="text-decoration: none; color: {self.color}" href="close://me.com">✖</a>'
+        self.setText(text)
+        self.resize(self.sizeHint())
+        self.position_in_parent()
+        self.show()
+        self.raise_()
+        self.close_timer.start(timeout)
+
+    def position_in_parent(self):
+        p = self.parent()
+        self.move((p.width() - self.width()) // 2, self.OFFSET_FROM_TOP)
+
+
+
 if __name__ == '__main__':
     from calibre.gui2 import Application
     app = Application([])
