@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 
@@ -27,7 +26,6 @@ from calibre.ebooks.oeb.polish.container import Container as ContainerBase
 from calibre.ebooks.oeb.polish.cover import (
     find_cover_image, find_cover_image_in_page, find_cover_page
 )
-from calibre.ebooks.oeb.polish.pretty import pretty_script_or_style
 from calibre.ebooks.oeb.polish.toc import from_xpaths, get_landmarks, get_toc
 from calibre.ebooks.oeb.polish.utils import guess_type
 from calibre.ptempfile import PersistentTemporaryDirectory
@@ -194,7 +192,8 @@ def toc_anchor_map(toc):
         if name and node['id'] not in seen_map[name]:
             ans[name].append({'id':node['id'], 'frag':node['frag']})
             seen_map[name].add(node['id'])
-        tuple(map(process_node, node['children']))
+        for i in node['children']:
+            process_node(i)
 
     process_node(toc)
     return dict(ans)
@@ -334,7 +333,6 @@ def transform_inline_styles(container, name, transform_sheet, transform_style):
             if nraw != style.text:
                 changed = True
                 style.text = nraw
-                pretty_script_or_style(container, style)
     for elem in root.xpath('//*[@style]'):
         text = elem.get('style', None)
         if text:
@@ -349,6 +347,7 @@ def transform_html(container, name, virtualize_resources, link_uid, link_to_map,
     link_xpath = XPath('//h:a[@href]')
     svg_link_xpath = XPath('//svg:a')
     img_xpath = XPath('//h:img[@src]')
+    svg_img_xpath = XPath('//svg:image[@xl:href]')
     res_link_xpath = XPath('//h:link[@href]')
     root = container.parsed(name)
     changed_names = set()
@@ -359,6 +358,10 @@ def transform_html(container, name, virtualize_resources, link_uid, link_to_map,
         img_name = container.href_to_name(img.get('src'), name)
         if img_name:
             img.set('data-calibre-src', img_name)
+    for img in svg_img_xpath(root):
+        img_name = container.href_to_name(img.get(XLINK('href')), name)
+        if img_name:
+            img.set('data-calibre-src', img_name)
 
     # Disable non-stylesheet link tags. This link will not be loaded by the
     # browser anyway and will causes the resource load check to hang
@@ -392,6 +395,8 @@ def transform_html(container, name, virtualize_resources, link_uid, link_to_map,
             href = a.get(attr)
             if href:
                 href = link_replacer(name, href)
+            elif attr in a.attrib:
+                a.set(attr, 'javascript:void(0)')
             if href and href.startswith(link_uid):
                 a.set(attr, 'javascript:void(0)')
                 parts = href.split('|')
@@ -418,8 +423,8 @@ class RenderManager:
         self.max_workers = max_workers
 
     def launch_worker(self):
-        with lopen(os.path.join(self.tdir, '{}.json'.format(len(self.workers))), 'wb') as output:
-            error = lopen(os.path.join(self.tdir, '{}.error'.format(len(self.workers))), 'wb')
+        with open(os.path.join(self.tdir, f'{len(self.workers)}.json'), 'wb') as output:
+            error = open(os.path.join(self.tdir, f'{len(self.workers)}.error'), 'wb')
             p = start_pipe_worker('from calibre.srv.render_book import worker_main; worker_main()', stdout=error, stderr=error)
             p.output_path = output.name
             p.error_path = error.name
@@ -487,10 +492,10 @@ class RenderManager:
                 worker.wait()
                 continue
             if worker.wait() != 0:
-                with lopen(worker.error_path, 'rb') as f:
+                with open(worker.error_path, 'rb') as f:
                     error = f.read().decode('utf-8', 'replace')
             else:
-                with lopen(worker.output_path, 'rb') as f:
+                with open(worker.output_path, 'rb') as f:
                     results.append(msgpack_loads(f.read()))
         if error is not None:
             raise Exception('Render worker failed with error:\n' + error)
@@ -530,6 +535,8 @@ def virtualize_html(container, name, link_uid, link_to_map, virtualized_names):
         elif href:
             a.set('target', '_blank')
             a.set('rel', 'noopener noreferrer')
+        elif attr in a.attrib:
+            a.set(attr, 'javascript:void(0)')
 
     for a in link_xpath(root):
         handle_link(a)
@@ -700,7 +707,7 @@ def process_exploded_book(
             amap[k] = tuple(v)  # needed for JSON serialization
 
     data = as_bytes(json.dumps(book_render_data, ensure_ascii=False))
-    with lopen(os.path.join(container.root, 'calibre-book-manifest.json'), 'wb') as f:
+    with open(os.path.join(container.root, 'calibre-book-manifest.json'), 'wb') as f:
         f.write(data)
 
     return container, bookmark_data
@@ -786,7 +793,7 @@ def render(pathtoebook, output_dir, book_hash=None, serialize_metadata=False, ex
         if serialize_metadata:
             from calibre.customize.ui import quick_metadata
             from calibre.ebooks.metadata.meta import get_metadata
-            with lopen(pathtoebook, 'rb') as f, quick_metadata:
+            with open(pathtoebook, 'rb') as f, quick_metadata:
                 mi = get_metadata(f, os.path.splitext(pathtoebook)[1][1:].lower())
         book_fmt, opfpath, input_fmt = extract_book(pathtoebook, output_dir, log=default_log)
         container, bookmark_data = process_exploded_book(
@@ -799,14 +806,14 @@ def render(pathtoebook, output_dir, book_hash=None, serialize_metadata=False, ex
             d = metadata_as_dict(mi)
             d.pop('cover_data', None)
             serialize_datetimes(d), serialize_datetimes(d.get('user_metadata', {}))
-            with lopen(os.path.join(output_dir, 'calibre-book-metadata.json'), 'wb') as f:
+            with open(os.path.join(output_dir, 'calibre-book-metadata.json'), 'wb') as f:
                 f.write(json_dumps(d))
         if extract_annotations:
             annotations = None
             if bookmark_data:
                 annotations = json_dumps(tuple(get_stored_annotations(container, bookmark_data)))
             if annotations:
-                with lopen(os.path.join(output_dir, 'calibre-book-annotations.json'), 'wb') as f:
+                with open(os.path.join(output_dir, 'calibre-book-annotations.json'), 'wb') as f:
                     f.write(annotations)
 
 
